-- GLOBAL --
DA = DA or { };
-- MODULE --
local _G = getfenv(0);
setmetatable(DA, {__index = _G});
setfenv(1, DA);
-- LOCAL --
title		= "Dimmed Actions"
version		= "1.0.10"
titleFull	= title.." "..version

coreFrame = CreateFrame("Frame");
core = LibStub("AceAddon-3.0"):NewAddon(coreFrame, title, "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0") -- 

--Echo strings
local strDebugFrom		= "|cffffff00[%s]|r" --Yellow function name. help pinpoint where the debug msg is from.
local strWhiteBar		= "|cffffff00 || |r" -- a white bar to seperate the debug info.

-- API Lookups
local ActionHasRange		= ActionHasRange;
local BOOKTYPE_SPELL		= BOOKTYPE_SPELL;
local GetActionInfo			= GetActionInfo;
local GetMacroInfo_			= GetMacroInfo;
local GetSpellCooldown_		= GetSpellCooldown;
local GetSpellLink			= GetSpellLink;
local GetSpellName			= GetSpellName;
local GetTime_				= GetTime;
local HasAction				= HasAction;
local InCombatLockdown_		= InCombatLockdown;
local IsActionInRange		= IsActionInRange;
local IsAltKeyDown			= IsAltKeyDown;
local IsUsableAction		= IsUsableAction;
local SecureCmdOptionParse_	= SecureCmdOptionParse;
local UnitAffectingCombat	= UnitAffectingCombat;
local UnitBuff				= UnitBuff;
local UnitDebuff			= UnitDebuff;
local UnitExists			= UnitExists;
local UnitIsEnemy			= UnitIsEnemy;
local UnitIsFriend			= UnitIsFriend;
local pairs					= pairs;
local print					= print;
local table_getn			= table.getn;
local table_insert			= table.insert;
local table_sort			= table.sort;
local tostring				= tostring;
local tonumber				= tonumber;

-- Internal opts
local GCDspells = {"Shoot", "Throw", "Cure Poison", "Purify", "Purge"}--spells I can check to see if we're on global cooldown. Shoot for most classes, Cure Posion for druid, Purify for Pally, Purge for shammy.


--don't touch.
myTooltip = CreateFrame("GameTooltip", "DA_ToolTip", UIParent, "GameTooltipTemplate");
BT4	= false; --Is bartender running? this gets set in OnEnable()
spellNames = {};--spellname to IDs.
hookedButtons = {};
barOpts		= false; --needed for bar display options.
coreOpts	= false;
existsOpts	= false;
stackOpts	= false;

local regEvents = {
	"ACTIONBAR_SLOT_CHANGED",
--~ 	"COMBAT_LOG_EVENT_UNFILTERED", --only registered if BT4 is not running.
	"PLAYER_REGEN_DISABLED",
	"PLAYER_REGEN_ENABLED",
}
local targetSpells = {--spells that can be cast on other people.
	["the target"] = true,
	["the enemy"] = true,
	["enemy target"] = true,
	["friendly target"] = true,
	["friendly party"] = true,
}


--Default Settings
db				= {};
defaultSettings	= {profile = {}}
defaultSettings.profile.updateFrequency	= 0.25;
defaultSettings.profile.dimAlpha		= 0.2;--20%
defaultSettings.profile.showAlt			= true;
defaultSettings.profile.showNoTarget	= true;
defaultSettings.profile.dimCombatOnly	= true;
defaultSettings.profile.dimNoMana	= true;
defaultSettings.profile.dimOutOfRange	= true;
defaultSettings.profile.dimSpellExists	= true;
defaultSettings.profile.readMacros	= false;
defaultSettings.profile.macroFindSpellSpam	= false;
defaultSettings.profile.dimCooldown	= true;
defaultSettings.profile.healingSpellsOnTargetsTarget	= true;

--bars
defaultSettings.profile.barsBar1				= true;
defaultSettings.profile.barsBonusBar1			= true;
defaultSettings.profile.barsMultiBarBottomLeft	= false;--above your main action bar
defaultSettings.profile.barsMultiBarBottomRight	= false;--above your packs
defaultSettings.profile.barsMultiBarRight		= false;--vertical bar on right side of screen.
defaultSettings.profile.barsMultiBarLeft		= false;--left of ^ bar.

--BT4 Optinos
defaultSettings.profile.barsBT4Bar1		= true;
defaultSettings.profile.barsBT4Bar2		= false;
defaultSettings.profile.barsBT4Bar3		= false;
defaultSettings.profile.barsBT4Bar4		= false;
defaultSettings.profile.barsBT4Bar5		= false;
defaultSettings.profile.barsBT4Bar6		= false;
defaultSettings.profile.barsBT4Bar7		= false;
defaultSettings.profile.barsBT4Bar8		= false;
defaultSettings.profile.barsBT4Bar9		= false;
defaultSettings.profile.barsBT4Bar10	= false;

healSpells = {
	["Abolish Disease"] = true,
	["Abolish Poison"] = true,
	["Blessing of Protection"] = true,
	["Chain Heal"] = true,
	["Cleanse"] = true,
	["Cure Disease"] = true,
	["Cure Poison"] = true,
	["Dispel Magic"] = true,
	["First Aid"] = true,
	["Flash Heal"] = true,
	["Flash of Light"] = true,
	["Greater Heal"] = true,
	["Heal"] = true,
	["Healing Touch"] = true,
	["Healing Wave"] = true,
	["Holy Light"] = true,
	["Lesser Heal"] = true,
	["Lesser Healing Wave"] = true,
	["Power Word: Shield"] = true,
	["Purify"] = true,
	["Regrowth"] = true,
	["Rejuvenation"] = true,
	["Remove Curse"] = true,
	["Remove Lesser Curse"] = true,
	["Renew"] = true,
}

-- Ace3 Functions
function core:OnInitialize()
	db = LibStub("AceDB-3.0"):New("DA_DB", defaultSettings)
	CoreOptionsTable.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(db)--save option profile or load another chars opts.

	db.RegisterCallback(self, "OnProfileReset", "OnProfileReset")
	
	core:RegisterChatCommand("da", "MySlashProcessorFunc")
	
	local config = LibStub("AceConfig-3.0");
	local dialog = LibStub("AceConfigDialog-3.0");
	
	config:RegisterOptionsTable(title, CoreOptionsTable)
	coreOpts = dialog:AddToBlizOptions(title, "DA "..version)
	
	config:RegisterOptionsTable(title.."Bars", BarOptionsTable)
	barOpts = dialog:AddToBlizOptions(title.."Bars", "Bars", "DA "..version)
	
	config:RegisterOptionsTable(title.."SE", SpellExistOptionsTable)
	existsOpts = dialog:AddToBlizOptions(title.."SE", "Spell Exists", "DA "..version)
	
	config:RegisterOptionsTable(title.."SC", StackCountOptionsTable)
	
end


----------------------------------------------
function core:MySlashProcessorFunc(input)	--
-- /da function brings up the UI options.	--
----------------------------------------------
	InterfaceOptionsFrame_OpenToCategory(barOpts)
	InterfaceOptionsFrame_OpenToCategory(coreOpts)
end

function core:OnEnable()
	for i, event in pairs(regEvents) do 
		core:RegisterEvent(event);
	end
	
	if _G.Bartender4 then
		BT4 = true;
		AddBT4Options()
	else
		--Hook when the bonus bars show and hide so we hook bonus actions.
		if BonusActionBarFrame:HasScript("OnShow") and not core:IsHooked(BonusActionBarFrame, "OnShow") then
			core:HookScript(BonusActionBarFrame, "OnShow", function(self) 
				HookBonusActions()
			end);
		end

		if BonusActionBarFrame:HasScript("OnHide") and not core:IsHooked(BonusActionBarFrame, "OnHide") then
			core:HookScript(BonusActionBarFrame, "OnHide", function(self) 
				UnhookBonusActions()
			end);
		end
	end
	
	CollectSpellNameInfo()
	
	if db.profile.dimCombatOnly == false then
		HookActionButtons()
	end

end

----------------------------------------------------------------------
function core:OnProfileReset(...)									--
-- User has reset proflie, so we reset our spell exists options.	--
----------------------------------------------------------------------
	local whatHappen, info = select(1, ...);

	db.profile.se = nil;
	AddSpellExistOptions()
end


function core:OnDisable()
	-- I know when ace addons are disabled their hooks are released, these functions return the actions to 100% alpha.
	UnhookAllBars()
end

------------------------------
function GetMouseoverInfo()	--	/script DA.GetMouseoverInfo()
-- Used to debug.			--
------------------------------
	local self = GetMouseFocus();
	print(self:GetName().." - "..self.action)
	if self.action then
		local actionID = self.action
		
--~ 	name, iconTexture, body, local = GetMacroInfo(MacroID)
		local actionType, spellID = GetActionInfo(actionID)
		if actionType == "macro" then
			local body = GetMacroBody(spellID)
			local result, target = SecureCmdOptionParse_(body)
			Debug("GetMouseoverInfo 1", result, target)

		end
	end
end

--------------------------------------
function HookActionButtons()		--
-- Hook action button's OnUpdate.	--
--------------------------------------
	if BT4 == true then
		if db.profile.barsBT4Bar1 == true then
			CheckBar("BT4Bar1")
		end
		
		if db.profile.barsBT4Bar2 == true then
			CheckBar("BT4Bar2")
		end
		if db.profile.barsBT4Bar3 == true then
			CheckBar("BT4Bar3")
		end
		if db.profile.barsBT4Bar4 == true then
			CheckBar("BT4Bar4")
		end
		if db.profile.barsBT4Bar5 == true then
			CheckBar("BT4Bar5")
		end
		if db.profile.barsBT4Bar6 == true then
			CheckBar("BT4Bar6")
		end
		if db.profile.barsBT4Bar7 == true then
			CheckBar("BT4Bar7")
		end
		if db.profile.barsBT4Bar8 == true then
			CheckBar("BT4Bar8")
		end
		if db.profile.barsBT4Bar9 == true then
			CheckBar("BT4Bar9")
		end
		if db.profile.barsBT4Bar10 == true then
			CheckBar("BT4Bar10")
		end
	else
		if BonusActionBarFrame:IsShown() then
			HookBonusActions()
		else
			UnhookBonusActions()
		end
		
		if db.profile.barsMultiBarBottomRight == true then
			for i=1, NUM_ACTIONBAR_BUTTONS, 1 do
				HookActionButton(_G["MultiBarBottomRightButton"..i], "MultiBarBottomRight", i)
			end
		end
		
		if db.profile.barsMultiBarBottomLeft == true then
			for i=1, NUM_ACTIONBAR_BUTTONS, 1 do
				HookActionButton(_G["MultiBarBottomLeftButton"..i], "MultiBarBottomLeft", i)
			end
		end
		
		if db.profile.barsMultiBarRight == true then
			for i=1, NUM_ACTIONBAR_BUTTONS, 1 do
				HookActionButton(_G["MultiBarRightButton"..i], "MultiBarRight", i)
			end
		end
		if db.profile.barsMultiBarLeft == true then
			for i=1, NUM_ACTIONBAR_BUTTONS, 1 do
				HookActionButton(_G["MultiBarLeftButton"..i], "MultiBarLeft", i)
			end
		end
	end
end

------------------------------------------------------------------
function HookBonusActions()										--
--	Here we loop threw bonus action buttons to hook them later.	--
------------------------------------------------------------------
	if db.profile.barsBonusBar1 == true then
		for i=1, NUM_ACTIONBAR_BUTTONS, 1 do
			HookActionButton(_G["BonusActionButton"..i], "BonusBar1", i)
		end

		UnhookBar("Bar1")
		
		--I hide BonusActionBarTexture0 & BonusActionBarTexture1 so I need to hide the buttons underneath too.
		for i=1, NUM_ACTIONBAR_BUTTONS, 1 do
			_G["ActionButton"..i]:SetAlpha(0);
		end
	end

end

------------------------------------------------------
function UnhookBonusActions()						--
-- Loop threw bonus action buttons to onhook them.	--
------------------------------------------------------
	if db.profile.barsBar1 == true then
		for i=1, NUM_ACTIONBAR_BUTTONS, 1 do
			HookActionButton(_G["ActionButton"..i], "Bar1", i)
		end
		UnhookBar("BonusBar1")
	end
end

----------------------------------------------------------------------------------
function CheckBar(barName)														--
-- Used to hook Bartender4 bars, checks children for buttons then hooks them.	--
----------------------------------------------------------------------------------
	local bar = _G[barName];--"BT4Bar1"
	local kids = { bar:GetChildren() };
	for _, child in ipairs(kids) do
		if child:GetName():find("Button") then
			HookActionButton(child, barName, 0)
		end
	end
end

--------------------------------------------------
function HookActionButton(buttonFrame, bar, id)	--
-- Hook a action button's OnUpdate script.		--
--------------------------------------------------
	if buttonFrame then
		hookedButtons[bar] = hookedButtons[bar] or {};
		if id == 0 then
			id = table.getn(hookedButtons[bar]) + 1;
		end
		hookedButtons[bar][id] = buttonFrame
		
		if buttonFrame:HasScript("OnUpdate") and not core:IsHooked(buttonFrame, "OnUpdate") then
			buttonFrame.daLastUpdate = 0;
			core:HookScript(buttonFrame, "OnUpdate", ActionButton_OnUpdate);
		end
		
		if db.profile.macroFindSpellSpam == true then
			if buttonFrame.action then
				local actionID = buttonFrame.action;
				local actionType, spellID = GetActionInfo(actionID)
				if actionType == "macro" then
					local name, _, body = GetMacroInfo_(spellID)
					local result, target = SecureCmdOptionParse_(body)
					
					if result and result ~= "" and not result:find("\n") then
						if result:find("!") then
							result = result:gsub("!","")
						end
						if not spellNames[result] then
							print("[Dimmed Actions] Cannot find "..result.." on bars, macro '"..name.."' won't be dimmed.")
						end
					end
				end
			end
		end
	end
end

------------------------------------------
function UnhookBar(bar)					--
-- Onhook all actions used by a bar.	--
------------------------------------------
	if hookedButtons[bar] then
		for _, buttonFrame in pairs(hookedButtons[bar]) do 
			buttonFrame:SetAlpha(1);
			if core:IsHooked(buttonFrame, "OnUpdate") then
				core:Unhook(buttonFrame, "OnUpdate");
			end
		end
	end
end

--------------------------------------------------
function ActionButton_OnUpdate(self, elapsed)	--
-- Our OnUpdate hooked function. 				--
--------------------------------------------------
	self.daLastUpdate = self.daLastUpdate + elapsed;
	if self.daLastUpdate >= db.profile.updateFrequency then 
		self.daLastUpdate = 0;
		UpdateActionButton(self)
	end
end


------------------------------------------------------------------------------
function CollectSpellNameInfo()												--	/script DA.CollectSpellNameInfo()
-- I save actionID and spellID indexed by spellname to help with macros.	--
-- Since macros don't tell us their spellID or a actionID, I check the 		--
-- spellname to our list and use the known spellID and actionID.			--
------------------------------------------------------------------------------
	local actionID;
	for b=0, 9 do
		for s = 1, 12 do
			actionID = s+(b*12);
			SaveActionInfo(actionID)
		end
	end
	
	if db.profile.dimSpellExists == true then
		AddSpellExistOptions()
	end
	
	AddStackOptions()
end

------------------------------------------------------------------------------------------------------
function SaveActionInfo(actionID)																	--
-- Save a actionID's name, spellID, weather it can be cast on a target and if it has a cooldown.	--
------------------------------------------------------------------------------------------------------
	local actionType, spellID, subType, spellRank, spellName;
	if HasAction(actionID) then -- and ActionHasRange_(actionID)
		actionType, spellID, subType  = GetActionInfo(actionID)
		if actionType == "spell" and subType  == "spell" and spellID and spellID > 0 then
			spellName, spellRank = GetSpellName( spellID, BOOKTYPE_SPELL );
			if spellName  then --and not spellNames[spellName]
				local targetSpell, hasCooldown, canStack = GetSpellDetails(spellName, spellRank)
				spellNames[spellName] = {aID=actionID, sID=spellID, targetSpell=targetSpell, hasCooldown = false, canStack = canStack}
				if spellName ~= "Shoot" then
					spellNames[spellName].hasCooldown = hasCooldown;
				end
			end
		end
	end
end

--------------------------------------------------------------
function core:ACTIONBAR_SLOT_CHANGED(event, ...)			--
-- A slot has changed, we save the spell's new actionID.	--
--------------------------------------------------------------
	local actionID  = select(1, ...)-- ***
	if actionID == 0 then
		CollectSpellNameInfo()
		return;
	end
	
	for spellName, data in pairs(spellNames) do 
		if data.aID == actionID then
			spellNames[spellName].aID = nil;
			break;
		end
	end
	if HasAction(actionID) then -- and ActionHasRange_(actionID)
		actionType, spellID, subType  = GetActionInfo(actionID)
		if actionType == "spell" and subType  == "spell" and spellID and spellID > 0 then
			spellName = GetSpellName( spellID, BOOKTYPE_SPELL );
			if spellName and spellNames[spellName] then
				spellNames[spellName].aID = actionID;
			end
		end
	end
end

--~ function FlagIsPlayer(flags)
--~ 	if bit.band(flags, COMBATLOG_OBJECT_CONTROL_PLAYER) == COMBATLOG_OBJECT_CONTROL_PLAYER then
--~ 		return true;
--~ 	end
--~ 	return nil;
--~ end

----------------------------------------------------------------------
function core:PLAYER_REGEN_DISABLED(event, ...)						--
-- We've entered combat, check if we should hook action buttons.	--
----------------------------------------------------------------------
	if db.profile.dimCombatOnly == true then
		HookActionButtons()
	end
end

--------------------------------------------------
function core:PLAYER_REGEN_ENABLED(event, ...)	--
-- We've exited combat, unhook action buttons.	--
--------------------------------------------------
	if db.profile.dimCombatOnly == true then
		UnhookAllBars()
	end
end

----------------------------------------------------------------------
function UnhookAllBars()											--
-- Unhook all bars, this also returns action buttons to 100% alpha.	--
----------------------------------------------------------------------
	if BT4 == true then
		UnhookBar("BT4Bar1")
		UnhookBar("BT4Bar2")
		UnhookBar("BT4Bar3")
		UnhookBar("BT4Bar4")
		UnhookBar("BT4Bar5")
		UnhookBar("BT4Bar6")
		UnhookBar("BT4Bar7")
		UnhookBar("BT4Bar8")
		UnhookBar("BT4Bar9")
		UnhookBar("BT4Bar10")
	else
		UnhookBar("Bar1")
		UnhookBar("BonusBar1")
		UnhookBar("MultiBarBottomLeft")
		UnhookBar("MultiBarBottomRight")
		UnhookBar("MultiBarRight")
		UnhookBar("MultiBarLeft")
	end
end

--removing this, gonna hook BonusBarFrame's OnShow and OnHide instead.
--~ function core:COMBAT_LOG_EVENT_UNFILTERED(event, ...)
--~ 	local timestamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags  = select(1, ...)
--~ 	local _, _, eventPrefix, eventSuffix = eventType:find("(.-)_(.+)")
--~ 	if eventPrefix == "SPELL" then
--~ 		if FlagIsPlayer(srcFlags) then
--~ 			local spellID, spellName, spellSchool, auraType  = select(9, ...)
--~ 			spellName = spellName:lower();
--~ 			if eventSuffix == "AURA_REMOVED" then
--~ 				if spellName:find("form") or spellName:find("stealth") or spellName:find("stance") or spellName:find("prowl") then
--~ 					UnhookBonusActions()
--~ 				end
--~ 			elseif eventSuffix == "AURA_APPLIED" then
--~ 				if spellName:find("form") or spellName:find("stealth") or spellName:find("stance") or spellName:find("prowl") then
--~ 					HookBonusActions()
--~ 				end
--~ 			end
--~ 		end
--~ 	end
--~ end

------------------------------------------------------------------------------------------
function UpdateActionButton(self)														--
-- Our main OnUpdate function, here we check if we should show or dim a action button.	--
------------------------------------------------------------------------------------------
	local actionID = self.action

	if db.profile.showAlt == true then
		if IsAltKeyDown() then
			self:SetAlpha(1);
			return;
		end
	end
	
	if db.profile.showNoTarget == true then
		if not UnitExists("target") then
			self:SetAlpha(1);
			return;
		end
	end
	
	local targetID = "target";
	
	local actionType, spellID = GetActionInfo(actionID)
	if actionType == "macro" then
		if db.profile.readMacros == true then
			local body = GetMacroBody(spellID)
			local result, target = SecureCmdOptionParse_(body)
			
			if target and target ~= targetID then 
				targetID = target;
			end
			if result and result ~= "" then
				if result:find("!") then
					result = result:gsub("!","")
				end
				if spellNames[result] then
					actionID = spellNames[result].aID or actionID;
					actionType = "spell"
					spellID = spellNames[result].sID;
				else
					self:SetAlpha(1);
					return;
				end
			else
				self:SetAlpha(1);
				return;
			end
		else
			self:SetAlpha(1);
			return;
		end
	end

	local isUsable, notEnoughMana = IsUsableAction(actionID)
	if db.profile.dimNoMana == true then
		if notEnoughMana then
			self:SetAlpha(db.profile.dimAlpha);
			return;
		end
	end

	if db.profile.dimOutOfRange == true then
		if ActionHasRange(actionID) and IsActionInRange(actionID,targetID) ~= 1 then
			self:SetAlpha(db.profile.dimAlpha);
			return;
		end
	end
	if actionType == "spell" then
		local spellName = GetSpellName( spellID, BOOKTYPE_SPELL );
		if db.profile.dimCooldown == true and (spellNames[spellName] and spellNames[spellName].hasCooldown == true) then
			local start, duration, enabled = GetSpellCooldown_(spellName);
			if start and start > 0 then --and ( start > 0 and duration > 0) then
				if start ~= GetGCD() then
					local wait = start + duration - GetTime_();
--~ 					Debug(spellName, start, duration, wait)
					if wait > 1 then--If there's less then 1 second remaining in the cooldown, show the action. This gives user a second warning to deside if they'll start mashing this next.
						self:SetAlpha(db.profile.dimAlpha);
						return;
					end
				end
			end
		end
		
		if db.profile.dimSpellExists == true and UnitExists(targetID) then
			if db.profile.se[spellName] == true then
				if db.profile.healingSpellsOnTargetsTarget == true and healSpells[spellName] then
					if not UnitIsFriend("player",targetID) then
						if UnitExists("targettarget") then
							targetID = "targettarget";
						else
							targetID = "player";
						end
					end
				end

				if IsBuffOn(spellName, targetID) then
					self:SetAlpha(db.profile.dimAlpha);
					return;
				end
			end

		end
	end

	self:SetAlpha(1);
end

--------------------------------------------------------------------------
function GetGCD()														--
-- Get Shoot's current cooldown to see if we're under global cooldown.	--
--------------------------------------------------------------------------
	for i=1, table.getn(GCDspells) do 
		if spellNames[GCDspells[i]] then
			local start, duration, enabled = GetSpellCooldown_(GCDspells[i]);
			if start and ( start > 0 and duration > 0) then
--~ 				local wait = start + duration - GetTime_();
--~ 				Debug("GetGCD", start, duration, wait)
--~ 				return wait;
				return start;
			end
		end
	end
	return 0;
end

--	/script 

--------------------------------------------------
function IsBuffOn(spellName, unitID)			--	/script DA.IsBuffOn("Lifebloom", "player")
-- Check if a spell is on a unitID.				--
-- If the unitID's friendly we check buffs.		--
-- If the unitID's hostile we check debuffs.	--
--------------------------------------------------
	local name, unitCaster, count;
	if UnitIsFriend("player", unitID) then 
		for i = 1, 40 do
			name, _, _, count, _, _, _, unitCaster = UnitBuff(unitID, i);
			if not name then break; end
			if name == spellName then
				if unitCaster and unitCaster == "player" then
					if spellNames[spellName].canStack == 0 or count >= db.profile.sc[spellName] then
						return true;
					end
				end
			end
		end
	else--if UnitIsEnemy_("player",unitID) then
		for i = 1, 40 do
			name, _, _, count, _, _, _, unitCaster = UnitDebuff(unitID, i)
			if not name then break; end
			if name == spellName then
				if unitCaster and unitCaster == "player" then
					if spellNames[spellName].canStack == 0 or count >= db.profile.sc[spellName] then
						return true;
					end
				end
			end
		end
	end
	return nil;
end

------------------------------
function Debug(from, ...)	--	/script DA.Debug("hey","bye")
-- simple print function.	--
------------------------------
	local tbl  = {...}--select(1, ...)
	local msg = tostring(tbl[1])
	for i=2,table_getn(tbl) do 
		msg = msg..strWhiteBar..tostring(tbl[i]);
	end
	print(strDebugFrom:format(from).." "..msg)
end

------------------------------------------------------------------------------------------------------
function GetSpellDetails(spellName, spellRank)														--
-- Check a spell's discription to see if spell is castable on a target and if it has a cooldown.	--
------------------------------------------------------------------------------------------------------
	if ( not myTooltip:IsShown() ) then
		myTooltip:SetOwner(UIParent, "ANCHOR_NONE");
	end
	local link = GetSpellLink(spellName, spellRank)
	if not link then
		return false;
	end
	
	local _, _, realid = link:find("Hspell:(%d+)")
	
	myTooltip:SetHyperlink("spell:"..realid)
	myTooltip:Show()

	local otherSpell = false;
	local hasCooldown = false;
	local canStack	= 0;
	local lines = myTooltip:NumLines()
	if lines > 0 then
		local qText = (_G["DA_ToolTipTextLeft"..lines]:GetText() or "")
		for tText in pairs(targetSpells) do 
			if qText:find("stack") or qText:find("applied") then
				for times in qText:gmatch("up to (.+) times") do
					canStack = tonumber(times or 0);
					break
				end
			end
			if (qText:find("over") or qText:find("every") or qText:find("sec")) and qText:find(tText) then
				otherSpell = true;
				break;
			end
		end

		if lines >= 3 then
			for i=1, lines do 
				local qText = (_G["DA_ToolTipTextRight"..i]:GetText() or "")
				if qText:find("cooldown") then
					hasCooldown = true;
					break;
				end
			end

		end
	end
	return otherSpell, hasCooldown, canStack;
end

------------------------------------------------------------------
function AddStackOptions() 										--
-- Add options of user has a spell that can stack on a target.	--
------------------------------------------------------------------
	local spells = {};
	for sN, data in pairs(spellNames) do 
		if data.canStack > 0 then
			table_insert(spells, sN);
		end
	end
	if table.getn(spells) > 0 then
		local dialog = LibStub("AceConfigDialog-3.0");
		if stackOpts == false then
			stackOpts = dialog:AddToBlizOptions(title.."SC", "Stack Count", "DA "..version)
		end
		
		table_sort(spells, function(a,b) 
			if(a and b) then 
				return a < b;
			end 
		end)
		
		defaultSettings.profile.sc = defaultSettings.profile.sc or {};--stack count
		db.profile.sc = db.profile.sc or {};
		
		for opt in pairs(StackCountOptionsTable.args) do 
			if opt ~= "what" then
				StackCountOptionsTable.args[opt] = nil;
			end
		end
		
		local index = 2;
		local spellName;
		for i=1, table.getn(spells) do 
			spellName = spells[i];
	
			defaultSettings.profile.sc[spellName] = spellNames[spellName].canStack;
			if db.profile.sc[spellName] == nil then
				db.profile.sc[spellName] = spellNames[spellName].canStack;
			end
			StackCountOptionsTable.args[spellName] = {
				type 		= "range",	 order		= index,
				name		= spellName,
				desc 		= "|cff00ff00How many times "..spellName.." should stack before considering it existing on your target.|r",
				min			= 1,
				max			= spellNames[spellName].canStack,
				step		= 1,
				set = function(info,val) 
					db.profile.sc[info[1]] = val;
				end,
				get = function(info) return db.profile.sc[info[1]] end
			}

			index = index + 1;
		end
		
	end
end

------------------------------------------------------------------
function AddSpellExistOptions()									--	/script DA.AddSpellExistOptions()
-- Add options to Spell Exists panel for every spell user has.	--
------------------------------------------------------------------
	local spells = {};
	for sN in pairs(spellNames) do 
		table_insert(spells, sN);
	end
	table_sort(spells, function(a,b) 
		if(a and b) then 
			return a < b;
		end 
	end)

	defaultSettings.profile.se = defaultSettings.profile.se or {};
	
	db.profile = db.profile or {};
	db.profile.se = db.profile.se or {};

	for opt in pairs(StackCountOptionsTable.args) do 
		if opt ~= "what" then
			StackCountOptionsTable.args[opt] = nil;
		end
	end
	
	local index = 2;
	local spellName;
	for i=1, table.getn(spells) do 
		spellName = spells[i];

		defaultSettings.profile.se[spellName] = spellNames[spellName].targetSpell;
		if db.profile.se[spellName] == nil then
			db.profile.se[spellName] = spellNames[spellName].targetSpell;
		end
		
		SpellExistOptionsTable.args[spellName] = {
			type = "toggle",	order	= index,
		
			name = spellName,
			desc = "|cff00ff00Check if "..spellName.." exists on target.|r",
		
			set = function(info,val) 
				db.profile.se[info[1]] = val;
			end,
			get = function(info) return db.profile.se[info[1]] end
		}
		
		index = index + 1;
	end
	
end

----------------------------------------------------------------------------------
function AddBT4Options()														--
-- User has Bartender4 running, lets add options to choose which bars to hook.	--
----------------------------------------------------------------------------------
	for opt in pairs(BarOptionsTable.args.defaultBars.args) do 
		BarOptionsTable.args.defaultBars.args[opt].disabled = true;
	end
	
	BarOptionsTable.args.BT4Bars = {
		name = "Bartender Bars",
		type = "group",
		order = 2,
		args={}
	}
	
	BarOptionsTable.args.BT4Bars.args.barsBT4Bar1 = {
		type = "toggle",	order	= 1,
	
		name = "Bar 1",
		desc = "|cff00ff00Work on Bartender bar 1|r",
	
		set = function(info,val) 
			db.profile.barsBT4Bar1 = val;
			if val == true then
				HookActionButtons()
			else
				UnhookBar("BT4Bar1")
			end

		end,
		get = function(info) return db.profile.barsBT4Bar1 end
	}

	BarOptionsTable.args.BT4Bars.args.barsBT4Bar2 = {
		type = "toggle",	order	= 2,
	
		name = "Bar 2",
		desc = "|cff00ff00Work on Bartender bar 2|r",
	
		set = function(info,val) 
			db.profile.barsBT4Bar2 = val;
			if val == true then
				HookActionButtons()
			else
				UnhookBar("BT4Bar2")
			end
		end,
		get = function(info) return db.profile.barsBT4Bar2 end
	}
	
	BarOptionsTable.args.BT4Bars.args.barsBT4Bar3 = {
		type = "toggle",	order	= 3,
	
		name = "Bar 3",
		desc = "|cff00ff00Work on Bartender bar 3|r",
	
		set = function(info,val) 
			db.profile.barsBT4Bar3 = val;
			if val == true then
				HookActionButtons()
			else
				UnhookBar("BT4Bar3")
			end
		end,
		get = function(info) return db.profile.barsBT4Bar3 end
	}
	BarOptionsTable.args.BT4Bars.args.barsBT4Bar4 = {
		type = "toggle",	order	= 4,
	
		name = "Bar 4",
		desc = "|cff00ff00Work on Bartender bar 4|r",
	
		set = function(info,val) 
			db.profile.barsBT4Bar4 = val;
			if val == true then
				HookActionButtons()
			else
				UnhookBar("BT4Bar4")
			end
		end,
		get = function(info) return db.profile.barsBT4Bar4 end
	}
	BarOptionsTable.args.BT4Bars.args.barsBT4Bar5 = {
		type = "toggle",	order	= 5,
	
		name = "Bar 5",
		desc = "|cff00ff00Work on Bartender bar 5|r",
	
		set = function(info,val) 
			db.profile.barsBT4Bar5 = val;
			if val == true then
				HookActionButtons()
			else
				UnhookBar("BT4Bar5")
			end
		end,
		get = function(info) return db.profile.barsBT4Bar5 end
	}
	BarOptionsTable.args.BT4Bars.args.barsBT4Bar6 = {
		type = "toggle",	order	= 6,
	
		name = "Bar 6",
		desc = "|cff00ff00Work on Bartender bar 6|r",
	
		set = function(info,val) 
			db.profile.barsBT4Bar6 = val;
			if val == true then
				HookActionButtons()
			else
				UnhookBar("BT4Bar6")
			end
		end,
		get = function(info) return db.profile.barsBT4Bar6 end
	}
	BarOptionsTable.args.BT4Bars.args.barsBT4Bar7 = {
		type = "toggle",	order	= 7,
	
		name = "Bar 7",
		desc = "|cff00ff00Work on Bartender bar 7|r",
	
		set = function(info,val) 
			db.profile.barsBT4Bar7 = val;
			if val == true then
				HookActionButtons()
			else
				UnhookBar("BT4Bar7")
			end
		end,
		get = function(info) return db.profile.barsBT4Bar7 end
	}
	BarOptionsTable.args.BT4Bars.args.barsBT4Bar8 = {
		type = "toggle",	order	= 8,
	
		name = "Bar 8",
		desc = "|cff00ff00Work on Bartender bar 8|r",
	
		set = function(info,val) 
			db.profile.barsBT4Bar8 = val;
			if val == true then
				HookActionButtons()
			else
				UnhookBar("BT4Bar8")
			end		end,
		get = function(info) return db.profile.barsBT4Bar8 end
	}
	BarOptionsTable.args.BT4Bars.args.barsBT4Bar9 = {
		type = "toggle",	order	= 9,
	
		name = "Bar 9",
		desc = "|cff00ff00Work on Bartender bar 9|r",
	
		set = function(info,val) 
			db.profile.barsBT4Bar9 = val;
			if val == true then
				HookActionButtons()
			else
				UnhookBar("BT4Bar9")
			end
		end,
		get = function(info) return db.profile.barsBT4Bar9 end
	}
	BarOptionsTable.args.BT4Bars.args.barsBT4Bar10 = {
		type = "toggle",	order	= 10,
	
		name = "Bar 10",
		desc = "|cff00ff00Work on Bartender bar 10|r",
	
		set = function(info,val) 
			db.profile.barsBT4Bar10 = val;
			if val == true then
				HookActionButtons()
			else
				UnhookBar("BT4Bar10")
			end
		end,
		get = function(info) return db.profile.barsBT4Bar10 end
	}
end

--------------------------
-- GUI Options Table.	--
--------------------------

StackCountOptionsTable = {
	name = "Stack Count",
	type = "group",
	childGroups = "tab",
	args = {},
}

StackCountOptionsTable.args.what = {
	type = "description",
	name = "How many times a buff/debuff should stack before considering it existing on your target.",
	width = "full",
	order = 1,
}

CoreOptionsTable = {
	name = title,
	type = "group",
	childGroups = "tab",

	args = {
		core={
			name = "Core",
			type = "group",
			order = 1,
			args={}
		},
	},
}

CoreOptionsTable.args.core.args.enable = {
	type = "toggle",	order	= 1,
	name	= "Enable",
	desc	= "Enables / Disables the addon",
	set = function(info,val) 
		if val == true then
			core:Enable();
		else
			core:Disable();
		end
	end,
	get = function(info) return core:IsEnabled() end
}

CoreOptionsTable.args.core.args.updateFrequency = {
	type 		= "range",	 order		= 2,
	name 		= "Update Frequency",
	desc 		= "|cff00ff00Delay between each action update.|r",
	min			= 0.1,
	max			= 1,
	step		= 0.05,
	set = function(info,val) 
		db.profile.updateFrequency = val;
	end,
	get = function(info) return db.profile.updateFrequency end
}

CoreOptionsTable.args.core.args.dimAlpha = {
	type 		= "range",	 order		= 3,
	name 		= "Dim Alpha",
	desc 		= "|cff00ff00How dim you wan't actions to become.|r",
	min			= 0.0,
	max			= 1, -- 10%
	step		= 0.05,
	isPercent	= true,
	set = function(info,val) 
		db.profile.dimAlpha = val;
	end,
	get = function(info) return db.profile.dimAlpha end
}


CoreOptionsTable.args.core.args.showAlt = {
	type = "toggle",	order	= 4,

	name = "Show Alt-Keypress",
	desc = "|cff00ff00Show all actions when pressing Alt.|r",

	set = function(info,val) 
		db.profile.showAlt = val;
	end,
	get = function(info) return db.profile.showAlt end
}

CoreOptionsTable.args.core.args.showNoTarget = {
	type = "toggle",	order	= 4,

	name = "Show No Target",
	desc = "|cff00ff00Show all actions when you have no target.|r",

	set = function(info,val) 
		db.profile.showNoTarget = val;
	end,
	get = function(info) return db.profile.showNoTarget end
}

CoreOptionsTable.args.core.args.dimCombatOnly = {
	type = "toggle",	order	= 5,

	name = "In Combat Only",
	desc = "|cff00ff00Only dim actions when in combat.|r",

	set = function(info,val) 
		db.profile.dimCombatOnly = val;
		
		if val == false then--enable when out combat, hook actions.
			HookActionButtons()
		elseif val == true and not InCombatLockdown_() then--enable only in combat and we're not in combat right now.
			UnhookAllBars()
		end
	end,
	get = function(info) return db.profile.dimCombatOnly end
}

CoreOptionsTable.args.core.args.dimNoMana = {
	type = "toggle",	order	= 6,

	name = "Dim No Mana",
	desc = "|cff00ff00Dim actions you don't have the mana/energy/rage for.|r",

	set = function(info,val) 
		db.profile.dimNoMana = val;
	end,
	get = function(info) return db.profile.dimNoMana end
}

CoreOptionsTable.args.core.args.dimOutOfRange = {
	type = "toggle",	order	= 7,

	name = "Dim out of range",
	desc = "|cff00ff00Dim actions that are out of range of your target.|r",

	set = function(info,val) 
		db.profile.dimOutOfRange = val;
	end,
	get = function(info) return db.profile.dimOutOfRange end
}

CoreOptionsTable.args.core.args.dimSpellExists = {
	type = "toggle",	order	= 8,

	name = "Dim Spell Exists",
	desc = "|cff00ff00Dim actions if spell exists on target.|r",

	set = function(info,val) 
		db.profile.dimSpellExists = val;
		
		if val == true then
			AddSpellExistOptions()
		else
			for opt in pairs(SpellExistOptionsTable.args) do 
				SpellExistOptionsTable.args[opt] = nil;
			end
		end
	end,
	get = function(info) return db.profile.dimSpellExists end
}
CoreOptionsTable.args.core.args.healingSpellsOnTargetsTarget = {
	type = "toggle",	order	= 9,

	name = "Healing Targets Target",
	desc = "|cff00ff00For healign spells, If your target's hostile, uses your target's target for Spell Exists.\n(Or self if targettarget doesn't exist)|r",

	set = function(info,val) 
		db.profile.healingSpellsOnTargetsTarget = val;
	end,
	get = function(info) return db.profile.healingSpellsOnTargetsTarget end
}


CoreOptionsTable.args.core.args.readMacros = {
	type = "toggle",	order	= 10,

	name = "Read Macros",
	desc = "|cff00ff00Attempt to dim macros.\nPlease put any spells the macro calls on your actionbar (doesn't matter which bar)|r",

	set = function(info,val) 
		db.profile.readMacros = val;
	end,
	get = function(info) return db.profile.readMacros end
}

CoreOptionsTable.args.core.args.macroFindSpellSpam = {
	type = "toggle",	order	= 11,

	name = "Missing Macro Spells",
	desc = "|cff00ff00For DA to work on macros, the spells within the macros need to be on a bar somewhere. Enable this and disable/enable the addon to see which spells are missing.|r",

	set = function(info,val) 
		db.profile.macroFindSpellSpam = val;
	end,
	get = function(info) return db.profile.macroFindSpellSpam end
}


CoreOptionsTable.args.core.args.dimCooldown = {
	type = "toggle",	order	= 12,

	name = "Dim Cooldown",
	desc = "|cff00ff00Dim spells on cooldown.|r",

	set = function(info,val) 
		db.profile.dimCooldown = val;
	end,
	get = function(info) return db.profile.dimCooldown end
}


--Bars
-- GUI settings
BarOptionsTable = {
	name = title,
	type = "group",
	childGroups = "tab",

	args = {
		defaultBars={
			name = "Blizzard Bars",
			type = "group",
			order = 1,
			args={}
		},
	},
}


BarOptionsTable.args.defaultBars.args.barsBar1 = {
	type = "toggle",	order	= 1,

	name = "Action Bar 1",
	desc = "|cff00ff00Work on the default action bar.|r",

	set = function(info,val) 
		db.profile.barsBar1 = val;
		if val == true then
			HookActionButtons()
		else
			UnhookBar("Bar1")
		end
	end,
	get = function(info) return db.profile.barsBar1 end
}
BarOptionsTable.args.defaultBars.args.barsBonusBar1 = {
	type = "toggle",	order	= 2,

	name = "Bonus Bar",
	desc = "|cff00ff00Work on the default bonus bar. (used when stealthing)|r",

	set = function(info,val) 
		db.profile.barsBonusBar1 = val;
		if val == true then
			HookActionButtons()
		else
			UnhookBar("BonusBar1")
		end
	end,
	get = function(info) return db.profile.barsBonusBar1 end
}


BarOptionsTable.args.defaultBars.args.barsMultiBarBottomLeft = {
	type = "toggle",	order	= 2,

	name = "Multi Bar Bottom Left",
	desc = "|cff00ff00Bars on your bottom left (above your main action bar)|r",

	set = function(info,val) 
		db.profile.barsMultiBarBottomLeft = val;
		if val == true then
			HookActionButtons()
		else
			UnhookBar("MultiBarBottomLeft")
		end
	end,
	get = function(info) return db.profile.barsMultiBarBottomLeft end
}

BarOptionsTable.args.defaultBars.args.barsMultiBarBottomRight = {
	type = "toggle",	order	= 3,

	name = "Multi Bar Bottom Right",
	desc = "|cff00ff00Bars on your bottom right (above your bags)|r",

	set = function(info,val) 
		db.profile.barsMultiBarBottomRight = val;
		if val == true then
			HookActionButtons()
		else
			UnhookBar("MultiBarBottomRight")
		end
	end,
	get = function(info) return db.profile.barsMultiBarBottomRight end
}



BarOptionsTable.args.defaultBars.args.barsMultiBarRight = {
	type = "toggle",	order	= 4,

	name = "Right Bar",
	desc = "|cff00ff00Bar on the right side of your screen.|r",

	set = function(info,val) 
		db.profile.barsMultiBarRight = val;
		if val == true then
			HookActionButtons()
		else
			UnhookBar("MultiBarRight")
		end
	end,
	get = function(info) return db.profile.barsMultiBarRight end
}

BarOptionsTable.args.defaultBars.args.barsMultiBarLeft = {
	type = "toggle",	order	= 5,

	name = "Right Bar 2",
	desc = "|cff00ff002nd bar on the right side of your screen.|r",

	set = function(info,val) 
		db.profile.barsMultiBarLeft = val;
		if val == true then
			HookActionButtons()
		else
			UnhookBar("MultiBarLeft")
		end
	end,
	get = function(info) return db.profile.barsMultiBarLeft end
}

SpellExistOptionsTable = {
	name = "Spell Exists",
	type = "group",
	childGroups = "tab",

	args = {},
}

SpellExistOptionsTable.args.what = {
	type = "description",
	name = "DA guesses which spells you want to check for on your target but it's not 100% perfect, select any extra dots/hots you want to look for on your target.",
	width = "full",
	order = 1,
}
